Method: TestQueue::Runner#queue_status

Defined in:
lib/test_queue/runner.rb

#queue_status(start_time, queue_size, local_worker_count, remote_worker_count) ⇒ Object

Subclasses can override to monitor the status of the queue.

For example, you may want to record metrics about how quickly remote workers connect, or abort the build if not enough connect.

This method is called very frequently during the test run, so don’t do anything expensive/blocking.

This method is not called on remote masters when using remote workers, only on the central master.

start_time - Time when the test run began queue_size - Integer number of suites left in the queue local_worker_count - Integer number of active local workers remote_worker_count - Integer number of active remote workers

Returns nothing.



625
626
# File 'lib/test_queue/runner.rb', line 625

def queue_status(start_time, queue_size, local_worker_count, remote_worker_count)
end